==> courier-imap-1.2.3.tar.gz: courier-imap-1.2.3/imap/pop3dserver.c <==
/*
** Copyright 1998 - 2000 Double Precision, Inc.
** See COPYING for distribution information.
**
** $Id: pop3dserver.c,v 1.3 2000/11/15 17:52:21 mrsam Exp $
*/

#if	HAVE_CONFIG_H
#include	"config.h"
#endif
#if HAVE_DIRENT_H
#include <dirent.h>
#define NAMLEN(dirent) strlen((dirent)->d_name)
#else
#define dirent direct
#define NAMLEN(dirent) (dirent)->d_namlen
#if HAVE_SYS_NDIR_H
#include <sys/ndir.h>
#endif
#if HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif
#if HAVE_NDIR_H
#include <ndir.h>
#endif
#endif
#include	<sys/types.h>
#if	HAVE_SYS_STAT_H
#include	<sys/stat.h>
#endif
#if TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#include <time.h>
#endif
#endif
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>
#if	HAVE_UNISTD_H
#include	<unistd.h>
#endif
#include	<signal.h>
#include	<errno.h>
#include	"authlib/authmod.h"
#include	"numlib/numlib.h"
#include	"maildir/config.h"
#include	"maildir/maildirmisc.h"
#include	"maildir/maildirquota.h"
#include	"maildir/maildirgetquota.h"

extern void pop3dcapa();

static const char *authaddr, *remoteip;

struct msglist {
	struct msglist *next;
	char *filename;
	int isdeleted;
	off_t size;
	} ;

static struct msglist *msglist_l;
static struct msglist **msglist_a;
static unsigned msglist_cnt;

/*
** The RFC is pretty strict in stating that octet size must count the CR
** in the CRLF endofline.
*/

static void calcsize(struct msglist *m)
{
FILE	*f=fopen(m->filename, "r");
int	c, lastc;

	m->size=0;
	if (f == 0)
	{
		perror("pop3d");
		return;
	}
	lastc='\n';
	while ((c=getc(f)) >= 0)
	{
		if (c == '\n')	++m->size;
		++m->size;
		lastc=c;
	}
	if (lastc != '\n')	m->size += 2;

	if (ferror(f))
	{
		perror("pop3d");
		return;
	}
	fclose(f);
}

/* Scan cur, and pick up all messages contained therein */

static void scancur()
{
DIR	*dirp;
struct	dirent *de;
struct	msglist *m;

	if ((dirp=opendir("cur")) == 0)
	{
		perror("pop3d");
		return;
	}

	while ((de=readdir(dirp)) != 0)
	{
		if ( de->d_name[0] == '.' )	continue;

		if ((m=(struct msglist *)malloc(sizeof(struct msglist))) == 0)
		{
			perror("malloc");
			break;
		}
		if ((m->filename=(char *)malloc(strlen(de->d_name)+5)) == 0)
		{
			free( (char *)m);
			perror("malloc");
			break;
		}
		strcat(strcpy(m->filename, "cur/"), de->d_name);
		m->isdeleted=0;
		m->next=msglist_l;
		msglist_l=m;
		msglist_cnt++;
		calcsize(m);
	}
	closedir(dirp);
}

/*
** When sorting messages, sort on the arrival date - the first part of the
** name of the file in the maildir is the timestamp.
*/

static int cmpmsgs(const void *a, const void *b)
{
const char *aname=(*(struct msglist **)a)->filename;
const char *bname=(*(struct msglist **)b)->filename;
const char *ap=strchr(aname, '/');
const char *bp=strchr(bname, '/');
int	n=0;

	if (ap && bp)
	{
	long	na=atol(ap+1);
	long	nb=atol(bp+1);

		if (na < nb)	return (-1);
		if (na > nb)	return (1);
		n=strcmp(ap, bp);
	}
	if (n == 0)	n=strcmp(aname, bname);
	return (n);
}

static void sortmsgs()
{
unsigned i;
struct msglist *m;

	if (msglist_cnt == 0)	return;

	if ((msglist_a=(struct msglist **)malloc(
			msglist_cnt*sizeof(struct msglist *))) == 0)
	{
		perror("pop3d");
		msglist_cnt=0;
		return;
	}

	for (i=0, m=msglist_l; m; m=m->next, i++)
		msglist_a[i]=m;
	qsort(msglist_a, msglist_cnt, sizeof(*msglist_a), cmpmsgs);
}

static void mkupper(char *p)
{
	while (*p)
	{
		*p=toupper(*p);
		p++;
	}
}

static void cleanup()
{
	unsigned i;
	char quotabuf[QUOTABUFSIZE];
	int has_quota=maildir_getquota(".", quotabuf);

	long deleted_bytes=0;
	long deleted_messages=0;

	for (i=0; i<msglist_cnt; i++)
		if (msglist_a[i]->isdeleted)
		{
			unsigned long un=0;

			const char *filename=msglist_a[i]->filename;

			if (has_quota == 0 && !MAILDIR_DELETED(filename))
			{
				if (maildir_parsequota(filename, &un))
				{
					struct stat stat_buf;

					if (stat(filename, &stat_buf) == 0)
						un=stat_buf.st_size;
				}
			}

			if (unlink(msglist_a[i]->filename))
				un=0;

			if (un)
			{
				deleted_bytes -= un;
				deleted_messages -= 1;
			}
		}

	if (deleted_messages < 0)
	{
		int quotafd;

		if (maildir_checkquota(".", &quotafd, quotabuf, deleted_bytes,
				       deleted_messages) && errno != EAGAIN &&
		    deleted_bytes >= 0)
		{
			if (quotafd >= 0)
				close (quotafd);
		}
		else
		{
			maildir_addquota(".", quotafd, quotabuf,
					 deleted_bytes, deleted_messages);
			if (quotafd >= 0)
				close(quotafd);
		}
	}
	return;
}

/* POP3 STAT */

static void do_stat()
{
off_t	n=0;
unsigned i, j;
char	buf[NUMBUFSIZE];

	j=0;
	for (i=0; i<msglist_cnt; i++)
	{
		if (msglist_a[i]->isdeleted)	continue;
		n += msglist_a[i]->size;
		++j;
	}

	printf("+OK %u %s\r\n", j, str_off_t(n, buf));
	fflush(stdout);
}

static unsigned getmsgnum(const char *p)
{
unsigned i;

	if (!p || (i=atoi(p)) > msglist_cnt || i == 0 ||
		msglist_a[i-1]->isdeleted)
	{
		printf("-ERR Invalid message number.\r\n");
		fflush(stdout);
		return (0);
	}
	return (i);
}

/* POP3 LIST */

static void do_list(const char *msgnum)
{
unsigned i;
char	buf[NUMBUFSIZE];

	if (msgnum)
	{
		if ((i=getmsgnum(msgnum)) != 0)
		{
			printf("+OK %u %s\r\n", i,
				str_off_t(msglist_a[i-1]->size, buf));
			fflush(stdout);
		}
		return;
	}

	printf("+OK POP3 clients that break here, they violate STD53.\r\n");
	for (i=0; i<msglist_cnt; i++)
	{
		if (msglist_a[i]->isdeleted)	continue;
		printf("%u %s\r\n", i+1, str_off_t(msglist_a[i]->size, buf));
	}
	printf(".\r\n");
	fflush(stdout);
}

/* RETR and TOP POP3 commands */

static void do_retr(unsigned i, unsigned *lptr)
{
FILE	*f=fopen(msglist_a[i]->filename, "r");
char	*p;
int	c, lastc='\n';
int	inheader=1;
char	buf[NUMBUFSIZE];

	if (!f)
	{
		printf("-ERR Can't open the message file - it's gone!\r\n");
		fflush(stdout);
		return;
	}
	printf( (lptr ? "+OK headers follow.\r\n":"+OK %s octets follow.\r\n"),
		str_off_t(msglist_a[i]->size, buf));

	for (lastc=0; (c=getc(f)) >= 0; lastc=c)
	{
		if (lastc == '\n')
		{
			if (lptr)
			{
				if (inheader)
				{
					if (c == '\n')	inheader=0;
				}
				else if ( *lptr-- == 0)	break;
			}

			if (c == '.')
				putchar('.');
		}
		if (c == '\n')	putchar('\r');
		putchar(c);
	}
	if (lastc != '\n')	printf("\r\n");
	printf(".\r\n");
	fflush(stdout);
	fclose(f);
	if (lptr)	return;

	if ((p=strchr(msglist_a[i]->filename, ':')) != 0 &&
		(p[1] != '2' || p[2] != ',' || strchr(p, 'S') != 0))
		return;

	if ((p=malloc(strlen(msglist_a[i]->filename)+5)) == 0)
		return;

	strcpy(p, msglist_a[i]->filename);
	if (strchr(p, ':') == 0)	strcat(p, ":2,");
	strcat(p, "S");

	if (rename(msglist_a[i]->filename, p))
	{
		free(p);
		return;
	}
	free(msglist_a[i]->filename);
	msglist_a[i]->filename=p;
}

/*
** The UIDL of the message is really just its filename, up to the first :
*/

static void print_uidl(unsigned i)
{
const char *p=strchr(msglist_a[i]->filename, '/')+1;

	while (*p && *p != ':')
	{
		if (*p < 0x21 || *p > 0x7E || *p == '\'' || *p == '"' ||
			*p == '+')
			printf("+%02X", (int)(unsigned char)*p);
		else
			putchar(*p);
		++p;
	}
	printf("\r\n");
}

static void do_uidl(const char *msgnum)
{
unsigned i;

	if (msgnum)
	{
		if ((i=getmsgnum(msgnum)) != 0)
		{
			printf("+OK %u ", i);
			print_uidl(i-1);
			fflush(stdout);
		}
		return;
	}
	printf("+OK\r\n");
	for (i=0; i<msglist_cnt; i++)
	{
		if (msglist_a[i]->isdeleted)	continue;
		printf("%u ", i+1);
		print_uidl(i);
	}
	printf(".\r\n");
	fflush(stdout);
}

static RETSIGTYPE bye(int signum)
{
static const char msg1[]="ERR: TIMEOUT, user=";
static const char msg2[]=", ip=[";
static const char msg3[]="]\n";
const char *p;

	write(2, msg1, sizeof(msg1)-1);
	p=authaddr;
	write(2, p, strlen(p));
	write(2, msg2, sizeof(msg2)-1);
	p=remoteip;
	write(2, p, strlen(p));
	write(2, msg3, sizeof(msg3)-1);
	exit(0);
#if	RETSIGTYPE != void
	return (0);
#endif
}

static void loop()
{
char	buf[BUFSIZ];
char	*p;
int	c;

	signal(SIGALRM, bye);
	while (alarm(300), fgets(buf, sizeof(buf), stdin))
	{
		alarm(0);
		if ((p=strchr(buf, '\n')) != 0)
			*p=0;
		else while ((c=getc(stdin)) >= 0 && c != '\n')
			;
		p=strtok(buf, " \t\r");
		if (!p)	p="";

		mkupper(p);
		if (strcmp(p, "QUIT") == 0)
		{
			printf("+OK Bye-bye.\r\n");
			fflush(stdout);
			cleanup();
			fprintf(stderr, "INFO: LOGOUT, user=%s, ip=[%s]\n",
					authaddr, remoteip);
			fflush(stderr);
			return;
		}

		if (strcmp(p, "STAT") == 0)
		{
			do_stat();
			continue;
		}

		if (strcmp(p, "LIST") == 0)
		{
			do_list(strtok(NULL, " \t\r"));
			continue;
		}

		if (strcmp(p, "RETR") == 0)
		{
		unsigned	i;

			if ((i=getmsgnum(strtok(NULL, " \t\r"))) == 0)
				goto error;

			do_retr(i-1, 0);
			continue;
		}

		if (strcmp(p, "CAPA") == 0)
		{
			pop3dcapa();
			continue;
		}

		if (strcmp(p, "DELE") == 0)
		{
		unsigned	i;

			if ((i=getmsgnum(strtok(NULL, " \t\r"))) == 0)
				goto error;

			msglist_a[i-1]->isdeleted=1;
			printf("+OK Deleted.\r\n");
			fflush(stdout);
			continue;
		}

		if (strcmp(p, "NOOP") == 0)
		{
			printf("+OK Yup.\r\n");
			fflush(stdout);
			continue;
		}

		if (strcmp(p, "RSET") == 0)
		{
		unsigned i;

			for (i=0; i<msglist_cnt; i++)
				msglist_a[i]->isdeleted=0;
			printf("+OK Resurrected.\r\n");
			fflush(stdout);
			continue;
		}

		if (strcmp(p, "TOP") == 0)
		{
		unsigned	i, j;
		const	char *q;

			if ((i=getmsgnum(strtok(NULL, " \t\r"))) == 0)
				goto error;

			q=strtok(NULL, " \t\r");

			if (!q)	goto error;

			j=atoi(q);
			do_retr(i-1, &j);
			continue;
		}

		if (strcmp(p, "UIDL") == 0)
		{
			do_uidl(strtok(NULL, " \t\r"));
			continue;
		}

error:
		printf("-ERR Invalid command.\r\n");
		fflush(stdout);
	}
	fprintf(stderr, "ERR: DISCONNECTED, user=%s, ip=[%s]\n",
			authaddr, remoteip);
	fflush(stderr);
}

/* Like every good Maildir reader, we purge the tmp subdirectory */

static void purgetmp()
{
DIR	*p=opendir("tmp");
time_t	t;
struct	dirent *de;
struct	stat	stat_buf;
char	*n;

	if (!p)	return;
	time (&t);
	t -= 48L * 60L * 60L;

	while ((de=readdir(p)) != 0)
	{
		if (de->d_name[0] == '.')	continue;
		n=malloc(strlen(de->d_name)+5);
		if (!n)	continue;
		strcat(strcpy(n, "tmp/"), de->d_name);
		if (stat(n, &stat_buf) == 0 && stat_buf.st_mtime < t)
			unlink(n);
		free(n);
	}
	closedir(p);
}

static void scannew()
{
DIR	*p=opendir("new");
struct	dirent *de;
char	*n, *o;

	if (!p)	return;

	while ((de=readdir(p)) != 0)
	{
		if (de->d_name[0] == '.')	continue;
		n=malloc(strlen(de->d_name)+5);
		if (!n)	continue;
		o=malloc(strlen(de->d_name)+8);
		if (!o)	continue;
		strcat(strcpy(n, "new/"), de->d_name);
		strcat(strcat(strcpy(o, "cur/"), de->d_name), ":2,");
		rename(n, o);
		free(n);
		free(o);
	}
	closedir(p);
}

#include	<unistd.h>

int main(int argc, char **argv)
{
char	*p;

	authmodclient();
	if ((p=getenv("MAILDIR")) != 0 && *p)
	{
		if (chdir(p))
		{
			perror(argv[1]);
			exit(1);
		}
	}
	else if (argc > 1 && chdir(argv[1]))
	{
		perror(argv[1]);
		exit(1);
	}
	if ((authaddr=getenv("AUTHADDR")) == 0 ||
		(remoteip=getenv("TCPREMOTEIP")) == 0)
	{
		fprintf(stderr, "ERROR: Required environment variables not initialized.\n");
		fflush(stderr);
		exit (1);
	}

	fprintf(stderr, "INFO: LOGIN, user=%s, ip=[%s]\n",
			authaddr,
			remoteip);
	fflush(stderr);

	msglist_cnt=0;
	msglist_l=0;
	msglist_a=0;
	purgetmp();
	scannew();
	scancur();
	sortmsgs();
	printf("+OK logged in.\r\n");
	fflush(stdout);
	loop();
	return (0);
}

==> courier-imap-1.3.5.tar.gz: courier-imap-1.3.5/imap/pop3dserver.c <==
/*
** Copyright 1998 - 2001 Double Precision, Inc.
** See COPYING for distribution information.
**
** $Id: pop3dserver.c,v 1.6 2001/02/11 16:44:12 mrsam Exp $
*/

#if	HAVE_CONFIG_H
#include	"config.h"
#endif
#if HAVE_DIRENT_H
#include <dirent.h>
#define NAMLEN(dirent) strlen((dirent)->d_name)
#else
#define dirent direct
#define NAMLEN(dirent) (dirent)->d_namlen
#if HAVE_SYS_NDIR_H
#include <sys/ndir.h>
#endif
#if HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif
#if HAVE_NDIR_H
#include <ndir.h>
#endif
#endif
#include	<sys/types.h>
#if	HAVE_SYS_STAT_H
#include	<sys/stat.h>
#endif
#if TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#include <time.h>
#endif
#endif
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>
#if	HAVE_UNISTD_H
#include	<unistd.h>
#endif
#include	<signal.h>
#include	<errno.h>
#include	"authlib/authmod.h"
#include	"numlib/numlib.h"
#include	"maildir/config.h"
#include	"maildir/maildirmisc.h"
#include	"maildir/maildirquota.h"
#include	"maildir/maildirgetquota.h"

extern void pop3dcapa();

static const char *authaddr, *remoteip;

struct msglist {
	struct msglist *next;
	char *filename;
	int isdeleted;
	off_t size;
	} ;

static struct msglist *msglist_l;
static struct msglist **msglist_a;
static unsigned msglist_cnt;

static unsigned long top_count=0;
static unsigned long retr_count=0;

/*
** The RFC is pretty strict in stating that octet size must count the CR
** in the CRLF endofline.
*/

static void calcsize(struct msglist *m)
{
FILE	*f=fopen(m->filename, "r");
int	c, lastc;

	m->size=0;
	if (f == 0)
	{
		perror("pop3d");
		return;
	}
	lastc='\n';
	while ((c=getc(f)) >= 0)
	{
		if (c == '\n')	++m->size;
		++m->size;
		lastc=c;
	}
	if (lastc != '\n')	m->size += 2;

	if (ferror(f))
	{
		perror("pop3d");
		return;
	}
	fclose(f);
}

/* Scan cur, and pick up all messages contained therein */

static void scancur()
{
DIR	*dirp;
struct	dirent *de;
struct	msglist *m;

	if ((dirp=opendir("cur")) == 0)
	{
		perror("pop3d");
		return;
	}

	while ((de=readdir(dirp)) != 0)
	{
		if ( de->d_name[0] == '.' )	continue;

		if ((m=(struct msglist *)malloc(sizeof(struct msglist))) == 0)
		{
			perror("malloc");
			break;
		}
		if ((m->filename=(char *)malloc(strlen(de->d_name)+5)) == 0)
		{
			free( (char *)m);
			perror("malloc");
			break;
		}
		strcat(strcpy(m->filename, "cur/"), de->d_name);
		m->isdeleted=0;
		m->next=msglist_l;
		msglist_l=m;
		msglist_cnt++;
		calcsize(m);
	}
	closedir(dirp);
}

/*
** When sorting messages, sort on the arrival date - the first part of the
** name of the file in the maildir is the timestamp.
*/

static int cmpmsgs(const void *a, const void *b)
{
const char *aname=(*(struct msglist **)a)->filename;
const char *bname=(*(struct msglist **)b)->filename;
const char *ap=strchr(aname, '/');
const char *bp=strchr(bname, '/');
int	n=0;

	if (ap && bp)
	{
	long	na=atol(ap+1);
	long	nb=atol(bp+1);

		if (na < nb)	return (-1);
		if (na > nb)	return (1);
		n=strcmp(ap, bp);
	}
	if (n == 0)	n=strcmp(aname, bname);
	return (n);
}

static void sortmsgs()
{
unsigned i;
struct msglist *m;

	if (msglist_cnt == 0)	return;

	if ((msglist_a=(struct msglist **)malloc(
			msglist_cnt*sizeof(struct msglist *))) == 0)
	{
		perror("pop3d");
		msglist_cnt=0;
		return;
	}

	for (i=0, m=msglist_l; m; m=m->next, i++)
		msglist_a[i]=m;
	qsort(msglist_a, msglist_cnt, sizeof(*msglist_a), cmpmsgs);
}

static void mkupper(char *p)
{
	while (*p)
	{
		*p=toupper(*p);
		p++;
	}
}

static void cleanup()
{
	unsigned i;
	char quotabuf[QUOTABUFSIZE];
	int has_quota=maildir_getquota(".", quotabuf);

	long deleted_bytes=0;
	long deleted_messages=0;

	for (i=0; i<msglist_cnt; i++)
		if (msglist_a[i]->isdeleted)
		{
			unsigned long un=0;

			const char *filename=msglist_a[i]->filename;

			if (has_quota == 0 && !MAILDIR_DELETED(filename))
			{
				if (maildir_parsequota(filename, &un))
				{
					struct stat stat_buf;

					if (stat(filename, &stat_buf) == 0)
						un=stat_buf.st_size;
				}
			}

			if (unlink(msglist_a[i]->filename))
				un=0;

			if (un)
			{
				deleted_bytes -= un;
				deleted_messages -= 1;
			}
		}

	if (deleted_messages < 0)
	{
		int quotafd;

		if (maildir_checkquota(".", &quotafd, quotabuf, deleted_bytes,
				       deleted_messages) && errno != EAGAIN &&
		    deleted_bytes >= 0)
		{
			if (quotafd >= 0)
				close (quotafd);
		}
		else
		{
			maildir_addquota(".", quotafd, quotabuf,
					 deleted_bytes, deleted_messages);
			if (quotafd >= 0)
				close(quotafd);
		}
	}
	return;
}

/* POP3 STAT */

static void do_stat()
{
off_t	n=0;
unsigned i, j;
char	buf[NUMBUFSIZE];

	j=0;
	for (i=0; i<msglist_cnt; i++)
	{
		if (msglist_a[i]->isdeleted)	continue;
		n += msglist_a[i]->size;
		++j;
	}

	printf("+OK %u %s\r\n", j, str_off_t(n, buf));
	fflush(stdout);
}

static unsigned getmsgnum(const char *p)
{
unsigned i;

	if (!p || (i=atoi(p)) > msglist_cnt || i == 0 ||
		msglist_a[i-1]->isdeleted)
	{
		printf("-ERR Invalid message number.\r\n");
		fflush(stdout);
		return (0);
	}
	return (i);
}

/* POP3 LIST */

static void do_list(const char *msgnum)
{
unsigned i;
char	buf[NUMBUFSIZE];

	if (msgnum)
	{
		if ((i=getmsgnum(msgnum)) != 0)
		{
			printf("+OK %u %s\r\n", i,
				str_off_t(msglist_a[i-1]->size, buf));
			fflush(stdout);
		}
		return;
	}

	printf("+OK POP3 clients that break here, they violate STD53.\r\n");
	for (i=0; i<msglist_cnt; i++)
	{
		if (msglist_a[i]->isdeleted)	continue;
		printf("%u %s\r\n", i+1, str_off_t(msglist_a[i]->size, buf));
	}
	printf(".\r\n");
	fflush(stdout);
}

/* RETR and TOP POP3 commands */

static void do_retr(unsigned i, unsigned *lptr)
{
FILE	*f=fopen(msglist_a[i]->filename, "r");
char	*p;
int	c, lastc='\n';
int	inheader=1;
char	buf[NUMBUFSIZE];
unsigned long *cntr;

	if (!f)
	{
		printf("-ERR Can't open the message file - it's gone!\r\n");
		fflush(stdout);
		return;
	}
	printf( (lptr ? "+OK headers follow.\r\n":"+OK %s octets follow.\r\n"),
		str_off_t(msglist_a[i]->size, buf));

	cntr= &retr_count;
	if (lptr)
		cntr= &top_count;

	for (lastc=0; (c=getc(f)) >= 0; lastc=c)
	{
		if (lastc == '\n')
		{
			if (lptr)
			{
				if (inheader)
				{
					if (c == '\n')	inheader=0;
				}
				else if ( (*lptr)-- == 0)	break;
			}

			if (c == '.')
				putchar('.');
		}
		if (c == '\n')	putchar('\r');
		putchar(c);
		++*cntr;
	}
	if (lastc != '\n')	printf("\r\n");
	printf(".\r\n");
	fflush(stdout);
	fclose(f);
	if (lptr)	return;

	if ((p=strchr(msglist_a[i]->filename, ':')) != 0 &&
		(p[1] != '2' || p[2] != ',' || strchr(p, 'S') != 0))
		return;

	if ((p=malloc(strlen(msglist_a[i]->filename)+5)) == 0)
		return;

	strcpy(p, msglist_a[i]->filename);
	if (strchr(p, ':') == 0)	strcat(p, ":2,");
	strcat(p, "S");

	if (lptr	// Don't mark as seen for TOP
	    || rename(msglist_a[i]->filename, p))
	{
		free(p);
		return;
	}
	free(msglist_a[i]->filename);
	msglist_a[i]->filename=p;
}

/*
** The UIDL of the message is really just its filename, up to the first :
*/

static void print_uidl(unsigned i)
{
const char *p=strchr(msglist_a[i]->filename, '/')+1;

	while (*p && *p != ':')
	{
		if (*p < 0x21 || *p > 0x7E || *p == '\'' || *p == '"' ||
			*p == '+')
			printf("+%02X", (int)(unsigned char)*p);
		else
			putchar(*p);
		++p;
	}
	printf("\r\n");
}

static void do_uidl(const char *msgnum)
{
unsigned i;

	if (msgnum)
	{
		if ((i=getmsgnum(msgnum)) != 0)
		{
			printf("+OK %u ", i);
			print_uidl(i-1);
			fflush(stdout);
		}
		return;
	}
	printf("+OK\r\n");
	for (i=0; i<msglist_cnt; i++)
	{
		if (msglist_a[i]->isdeleted)	continue;
		printf("%u ", i+1);
		print_uidl(i);
	}
	printf(".\r\n");
	fflush(stdout);
}

static void acctout(const char *disc)
{
	static const char msg1[]="ERR: ";
	static const char msg2[]=", user=";
	static const char msg3[]=", ip=[";
	static const char msg4[]="], top=";
	static const char msg5[]=", retr=";
	char num1[NUMBUFSIZE];
	char num2[NUMBUFSIZE];
	char *p;

	str_size_t(top_count, num1);
	str_size_t(retr_count, num2);

	p=malloc(strlen(authaddr)+strlen(remoteip)+strlen(disc)+
		 strlen(num1)+strlen(num2)+200);	/* Should be enough */

	strcpy(p, msg1);
	strcat(p, disc);
	strcat(p, msg2);
	strcat(p, authaddr);
	strcat(p, msg3);
	strcat(p, remoteip);
	strcat(p, msg4);
	strcat(p, num1);
	strcat(p, msg5);
	strcat(p, num2);
	strcat(p, "\n");
	write(2, p, strlen(p));
	free(p);
}

static RETSIGTYPE bye(int signum)
{
	acctout("TIMEOUT");
	exit(0);
#if	RETSIGTYPE != void
	return (0);
#endif
}

static void loop()
{
char	buf[BUFSIZ];
char	*p;
int	c;

	signal(SIGALRM, bye);
	while (alarm(300), fgets(buf, sizeof(buf), stdin))
	{
		alarm(0);
		if ((p=strchr(buf, '\n')) != 0)
			*p=0;
		else while ((c=getc(stdin)) >= 0 && c != '\n')
			;
		p=strtok(buf, " \t\r");
		if (!p)	p="";

		mkupper(p);
		if (strcmp(p, "QUIT") == 0)
		{
			printf("+OK Bye-bye.\r\n");
			fflush(stdout);
			cleanup();
			acctout("LOGOUT");
			return;
		}

		if (strcmp(p, "STAT") == 0)
		{
			do_stat();
			continue;
		}

		if (strcmp(p, "LIST") == 0)
		{
			do_list(strtok(NULL, " \t\r"));
			continue;
		}

		if (strcmp(p, "RETR") == 0)
		{
		unsigned	i;

			if ((i=getmsgnum(strtok(NULL, " \t\r"))) == 0)
				continue;

			do_retr(i-1, 0);
			continue;
		}

		if (strcmp(p, "CAPA") == 0)
		{
			pop3dcapa();
			continue;
		}

		if (strcmp(p, "DELE") == 0)
		{
		unsigned	i;

			if ((i=getmsgnum(strtok(NULL, " \t\r"))) == 0)
				continue;

			msglist_a[i-1]->isdeleted=1;
			printf("+OK Deleted.\r\n");
			fflush(stdout);
			continue;
		}

		if (strcmp(p, "NOOP") == 0)
		{
			printf("+OK Yup.\r\n");
			fflush(stdout);
			continue;
		}

		if (strcmp(p, "RSET") == 0)
		{
		unsigned i;

			for (i=0; i<msglist_cnt; i++)
				msglist_a[i]->isdeleted=0;
			printf("+OK Resurrected.\r\n");
			fflush(stdout);
			continue;
		}

		if (strcmp(p, "TOP") == 0)
		{
		unsigned	i, j;
		const	char *q;

			if ((i=getmsgnum(strtok(NULL, " \t\r"))) == 0)
				continue;

			q=strtok(NULL, " \t\r");

			if (!q)	goto error;

			j=atoi(q);
			do_retr(i-1, &j);
			continue;
		}

		if (strcmp(p, "UIDL") == 0)
		{
			do_uidl(strtok(NULL, " \t\r"));
			continue;
		}

error:
		printf("-ERR Invalid command.\r\n");
		fflush(stdout);
	}
	acctout("DISCONNECTED");
}

/* Like every good Maildir reader, we purge the tmp subdirectory */

static void purgetmp()
{
DIR	*p=opendir("tmp");
time_t	t;
struct	dirent *de;
struct	stat	stat_buf;
char	*n;

	if (!p)	return;
	time (&t);
	t -= 48L * 60L * 60L;

	while ((de=readdir(p)) != 0)
	{
		if (de->d_name[0] == '.')	continue;
		n=malloc(strlen(de->d_name)+5);
		if (!n)	continue;
		strcat(strcpy(n, "tmp/"), de->d_name);
		if (stat(n, &stat_buf) == 0 && stat_buf.st_mtime < t)
			unlink(n);
		free(n);
	}
	closedir(p);
}

static void scannew()
{
DIR	*p=opendir("new");
struct	dirent *de;
char	*n, *o;

	if (!p)	return;

	while ((de=readdir(p)) != 0)
	{
		if (de->d_name[0] == '.')	continue;
		n=malloc(strlen(de->d_name)+5);
		if (!n)	continue;
		o=malloc(strlen(de->d_name)+8);
		if (!o)	continue;
		strcat(strcpy(n, "new/"), de->d_name);
		strcat(strcat(strcpy(o, "cur/"), de->d_name), ":2,");
		rename(n, o);
		free(n);
		free(o);
	}
	closedir(p);
}

#include	<unistd.h>

int main(int argc, char **argv)
{
char	*p;

	authmodclient();
	if ((p=getenv("MAILDIR")) != 0 && *p)
	{
		if (chdir(p))
		{
			perror(argv[1]);
			exit(1);
		}
	}
	else if (argc > 1 && chdir(argv[1]))
	{
		perror(argv[1]);
		exit(1);
	}
	if ((authaddr=getenv("AUTHADDR")) == 0 ||
		(remoteip=getenv("TCPREMOTEIP")) == 0)
	{
		fprintf(stderr, "ERROR: Required environment variables not initialized.\n");
		fflush(stderr);
		exit (1);
	}

	fprintf(stderr, "INFO: LOGIN, user=%s, ip=[%s]\n",
			authaddr,
			remoteip);
	fflush(stderr);

	msglist_cnt=0;
	msglist_l=0;
	msglist_a=0;
	purgetmp();
	scannew();
	scancur();
	sortmsgs();
	printf("+OK logged in.\r\n");
	fflush(stdout);
	loop();
	return (0);
}
